import os
import sys
import shutil
import glob
# this line is for when main.py is run directly
sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.abspath(__file__))))

from subs2cia.argparser import get_args_subs2cia
from subs2cia.sources import AVSFile, group_files
from subs2cia.condense import Condense
from subs2cia.CardExport import CardExport

from pathlib import Path
import logging
from pprint import pprint
from typing import Union, List

presets = [
    {  # preset 0
        'preset_description': "Padded and merged Japanese condensed audio",
        'threshold': 1500,
        'padding': 200,
        # 'partition_size': 1800,  # 30 minutes, for long movies
        'target_lang': 'ja',
    },
    {  # preset 1
        'preset_description': "Unpadded Japanese condensed audio",
        'threshold': 0,  # note: default is 0
        'padding': 0,  # note: default is 0
        # 'partition': 1800,  # 30 minutes, for long movies
        'target_lang': 'ja',
    },
]


def list_presets():
    for idx, preset in enumerate(presets):
        print(f"Preset {idx}")
        pprint(preset)


def condense_start(args, groups: List[List[AVSFile]]):
    condense_args = {key: args[key] for key in
                 ['outdir', 'outstem', 'condensed_video', 'padding', 'threshold', 'partition', 'split',
                  'demux_overwrite_existing', 'overwrite_existing_generated', 'keep_temporaries',
                  'target_lang', 'out_audioext', 'minimum_compression_ratio', 'use_all_subs', 'subtitle_regex_filter',
                  'subtitle_regex_substrfilter', 'subtitle_regex_substrfilter_nokeep',
                  'audio_stream_index', 'subtitle_stream_index', 'ignore_range', 'ignore_chapters',
                  'bitrate', 'mono_channel', 'interactive', 'no_condensed_subtitles']}

    condensed_files = [Condense(g, **condense_args) for g in groups]
    for c in condensed_files:
        c.get_and_partition_streams()
        c.initialize_pickers()
        if args['dry_run']:
            continue
        if args['list_streams']:
            c.list_streams()
            continue
        c.choose_streams()
        c.export()
        c.cleanup()


def srs_export_start(args, groups: List[List[AVSFile]]):
    srs_args = {key: args[key] for key in
                 ['outdir', 'outstem', 'condensed_video', 'padding', 'demux_overwrite_existing',
                  'overwrite_existing_generated', 'keep_temporaries', 'target_lang', 'out_audioext', 'use_all_subs',
                  'subtitle_regex_filter', 'audio_stream_index', 'subtitle_stream_index', 'ignore_range', 'ignore_chapters',
                  'bitrate', 'mono_channel', 'interactive', 'normalize_audio']
                }

    cardexport_group = [CardExport(g, **srs_args) for g in groups]
    for c in cardexport_group:
        c.get_and_partition_streams()
        c.initialize_pickers()
        if args['dry_run']:
            continue
        if args['list_streams']:
            c.list_streams()
            continue
        c.choose_streams()
        c.export()
        c.cleanup()


def start():
    if not shutil.which('ffmpeg'):
        logging.warning(f"Couldn't find ffmpeg in PATH, things may break.")

    args = get_args_subs2cia()
    args = vars(args)

    if args['verbose']:
        if args['debug']:
            logging.basicConfig(level=logging.DEBUG)
        else:
            logging.basicConfig(level=logging.INFO)
    elif args['debug']:
        logging.basicConfig(level=logging.DEBUG)

    from subs2cia import __version__
    logging.info(f"subs2cia version {__version__}")
    logging.debug(f"Start arguments: {args}")



    if args['list_presets']:
        list_presets()
        return

    if args['preset'] is not None:
        if abs(args['preset']) >= len(presets):
            logging.critical(f"Preset {args['preset']} does not exist")
            exit(0)
        logging.info(f"using preset {args['preset']}")
        for key, val in presets[args['preset']].items():
            if key in args.keys() and ((args[key] == False) or (args[key] is None)):  # override presets
                args[key] = val

    if args['infiles'] is None:
        logging.info("No input files given, nothing to do.")
        exit(0)

    infiles = _resolve(args['infiles'])

    if args['absolute_paths']:
        sources = [AVSFile(Path(file).absolute()) for file in infiles]
    else:
        sources = [AVSFile(Path(file)) for file in infiles]


    for s in sources:
        s.probe()
        s.get_type()

    if args['batch']:
        args['outstem'] = None
        logging.info(f"Running in batch mode, attempting to group input files together.")
        groups = list(group_files(sources))
    else:
        if len(sources) > 2:
            logging.warning(f"Redundant input files detected. Got {len(sources)} "
                            f"input files to process and not running "
                            f"in batch mode. Only one output "
                            f"will be generated. ")
        groups = [list(sources)]
    logging.info(f"Have {len(groups)} group(s) to process.")

    commands = {
        'condense': condense_start,
        'srs': srs_export_start
    }
    commands[args['command']](args, groups)

def _resolve(files):
    resolved = []

    for f in files:
        if '*' in f or '?' in f:
            globbed = glob.glob(f)
            if len(globbed) > 0:
                resolved += globbed
            else:
                resolved.append(f)
        else:
            resolved.append(f)

    return resolved

if __name__ == '__main__':
    start()



# each output has a set of possible input files
# all source_files does is take all of the input files and partitons them into three lists:

from subs2cia.ffmpeg_tools import ffmpeg_demux

from pathlib import Path
import logging
import ffmpeg
import pycountry
from typing import List, Union
from collections import defaultdict



class AVSFile:
    def __init__(self, filepath: Path):
        if not filepath.exists():
            raise AssertionError(f"File {filepath} does not exist")
        self.filepath = filepath
        self.info = None
        self.type = None

    # returns string-encoded type (subtitle, audio, video)
    # determining type may just be as simple as reading the extension
    # but sometimes its better to use a parser and make sure the extension is correct
    def probe(self):
        try:
            self.info = ffmpeg.probe(self.filepath, 'ffprobe', **{'show_chapters': None})
        except ffmpeg.Error as e:
            logging.warning(
                f"Couldn't probe file, skipping {str(self.filepath)}. ffmpeg output: \n" + e.stderr.decode("utf-8"))

        logging.debug(f"ffmpeg probe results: {self.info}")

    def get_type(self):
        if self.info is None:  # ffprobe probably failed for some reason
            self.type = 'unknown'
            return
        if 'streams' not in self.info:
            logging.warning(f"Unexpected ffmpeg.probe output, ignoring file {str(self.filepath)}")
            logging.debug(self.info)
            self.type = None
            return
        if len(self.info['streams']) > 1:
            self.type = 'video'  # video files are treated as multi-stream objects
            return
        stream = self.info['streams'][0]
        if stream['codec_type'] == 'video':
            logging.info(f"File {str(self.filepath)} contains no audio or subtitle tracks!")
        self.type = stream['codec_type']

    def __str__(self):
        return f"{str(self.filepath)} ({self.type})"

    def __repr__(self):
        return f"AVSFile(filepath={self.filepath.__repr__()})"


# single ffmpeg stream
class Stream:
    index = None

    def __init__(self, file: AVSFile, type, stream_info, index=None, ):
        self.file = file
        self.index = index
        self.type = type
        # index of None indicates that demuxing with ffmpeg is not nessecary to extract data
        self.demux_file = None
        self.lang = 'unknownlang'
        self.stream_info = stream_info

        self.get_language()

    def is_standalone(self):
        if self.index is None:
            return True
        return False

    def __str__(self):
        if self.is_standalone():
            return f"standalone {self.stream_info['codec_name']} {self.type} at {str(self.file)}"
        else:
            return f"stream {self.index} ({self.type}, {self.stream_info['codec_name']}) in {self.file}"

    def __repr__(self):
        return f"Stream(file={self.file.__repr__()}, type={self.type}, index={self.index})"

    def get_language(self):
        if self.lang != 'unknownlang':
            return self.lang.alpha_3
        if self.is_standalone():
            # no metadata to analyze, look in suffixes for language codes
            suffixes = self.file.filepath.suffixes
            if len(suffixes) >= 2:
                if suffixes[-2] != '.forced':
                    self.lang = suffixes[-2][1:]
                elif len(suffixes) != 2:
                    self.lang = suffixes[-3][1:]
            if self.lang == 'unknownlang':
                return self.lang
            try:
                lang = pycountry.languages.lookup(self.lang)
            except:
                logging.warning(f"{self.lang} is not a language, treating {self.file.filepath} as unknown language")
                self.lang = 'unknownlang'
                return self.lang
            self.lang = lang
            return self.lang.alpha_3
        # look at metadata for language codes
        if 'tags' not in self.file.info['streams'][self.index]:
            return self.lang
        if 'language' not in self.file.info['streams'][self.index]['tags']:
            return self.lang
        try:
            self.lang = pycountry.languages.lookup(self.file.info['streams'][self.index]['tags']['language'])
        except LookupError as e:
            logging.warning(f"{self} language {self.file.info['streams'][self.index]['tags']['language']} is not a "
                            f"proper language code, setting to unknown language.")
            self.lang = 'unknownlang'
            return self.lang
        return self.lang.alpha_3

    def demux(self, overwrite_existing: bool):
        if self.demux_file is not None:
            return self.demux_file
        demux_path = self.file.filepath
        if not self.is_standalone():
            if self.type == 'subtitle':
                subtitle_mapping = {
                    'subrip': 'srt',
                    'ass': 'ass'
                }
                # todo: bitmap subtitles
                if self.stream_info is not None and 'codec_name' in self.stream_info:
                    if self.stream_info['codec_name'] not in subtitle_mapping:
                        extension = 'ass'
                        logging.warning(f"Unknown subtitle type {self.stream_info['codec_name']} found, "
                                     f"will attempt to convert to .ass")
                    else:
                        extension = subtitle_mapping[self.stream_info['codec_name']]

            if self.type == 'audio':
                # we could change what type to demux as similarly to subtitles,
                # but it may cause compatability issues down the road so let's
                # keep it as flac for now
                extension = 'flac'
            demux_path = self.file.filepath.parent / Path(
                f'{self.file.filepath.name}.stream{self.index}.{self.type}.{self.get_language()}.{extension}')

            if overwrite_existing or not demux_path.exists() or demux_path.stat().st_size == 0:
                demux_path = ffmpeg_demux(self.file.filepath, self.index, demux_path)
                if demux_path is None:
                    logging.error(
                        f"Couldn't demux stream {self.index} from {str(self.file.filepath)} (type={self.type})")
                    return None
        self.demux_file = AVSFile(demux_path)
        self.demux_file.probe()
        self.demux_file.get_type()
        return self.demux_file

    def cleanup_demux(self):
        if self.demux_file is not None and self.index is not None:
            logging.info(f"Deleting temporary file {str(self.demux_file.filepath)}")
            self.demux_file.filepath.unlink()

    # return a readable path to the data
    def get_data_path(self) -> Path:
        if self.is_standalone() or self.type == 'video':
            return self.file.filepath
        else:
            return self.demux_file.filepath


def common_count(t0, t1):
    # returns the length of the longest common prefix
    i = 0
    for i, pair in enumerate(zip(t0, t1)):
        if pair[0] != pair[1]:
            return i
    return i + 1


def is_language(s):
    try:
        pycountry.languages.lookup(s)
        return True
    except:
        return False


# Strip extensions and language info from local assets, Plex style.
# see https://support.plex.tv/articles/200471133-adding-local-subtitles-to-your-media/
def strip_extensions(p: Path) -> Path:
    if len(p.suffixes) <= 1:
        return p.with_suffix('')
    p = p.with_suffix('')
    if len(p.suffixes) >= 1:
        if p.suffixes[-1] == '.forced':
            p = p.with_suffix('')
    if len(p.suffixes) >= 1:
        lcode = p.suffixes[-1]
        lcode = lcode.replace('.', '')
        if is_language(lcode):
            return p.with_suffix('')
        # return p
    return p


def group_names_better(sources: List[AVSFile]) -> List[List[AVSFile]]:
    all_groups = []
    while(len(sources) > 0):
        group = [sources.pop(0)]
        to_remove = []
        for f in sources:
            if strip_extensions(f.filepath) == strip_extensions(group[0].filepath):
                group.append(f)
                to_remove.append(f)
        for f in to_remove:
            sources.remove(f)
        all_groups.append(group)
    return all_groups


def group_files(sources: [AVSFile]):
    file_groups = group_names_better(sources)
    logging.debug(f"groups: {[[f.filepath for f in g] for g in file_groups]}")
    return file_groups


def get_and_partition_streams(sources: List[AVSFile]):
    partitioned_streams = defaultdict(list)
    for sourcefile in sources:
        if sourcefile.type == 'video':
            # dig into streams
            for idx, stream_info in enumerate(sourcefile.info['streams']):
                stype = stream_info['codec_type']
                partitioned_streams[stype].append(Stream(file=sourcefile, type=stype,
                                                              index=idx, stream_info=stream_info))
            continue
        partitioned_streams[sourcefile.type].append(Stream(file=sourcefile, type=sourcefile.type,
                                                                stream_info=sourcefile.info['streams'][0],
                                                                index=None))
        # for stream in sourcefile
    for k in partitioned_streams:
        logging.info(f"Found {len(partitioned_streams[k])} {k} input streams")
        # logging.debug(f"Streams found: {self.partitioned_streams[k]}")
    return partitioned_streams



import argparse
import sys
import re
from pysubs2.time import make_time


def get_args_subzipper():
    parser = argparse.ArgumentParser(description=f'SubZipper: Map video files to subtitle files')

    parser.add_argument('-s', '--subtitle', metavar='<input files>', dest='subfiles', default=None, required=True,
                        type=str, nargs='+',
                        help='List of subtitle files. Number of subtitle files should equal number of reference files.')

    parser.add_argument('-r', '--reference', metavar='<input files>', dest='reffiles', default=None, required=True,
                        type=str, nargs='+',
                        help='List of reference files, typically video files. '
                             'Number of subtitle files should equal number of reference files. ')

    parser.add_argument('-l', '--language', metavar='ISO_LANG_CODE', dest='lang', default=None, required=False,
                        type=str,
                        help='Language code to append to end of subtitle file. Optional. '
                             'If set, will be checked for validity.')

    parser.add_argument('-ns', '--no-sort', action='store_true', dest='no_sort', default=False,
                        help="If set, will not sort input files alphabetically.")

    parser.add_argument('-d', '--dry-run', action='store_true', dest='dry_run', default=False,
                        help="If set, will print out mappings but will not write any changes to disk.")

    parser.add_argument('-v', '--verbose', action='store_true', dest='verbose', default=False,
                        help='Verbose output if set.')

    # parser.add_argument('-vv', '--debug', action='store_true', dest='debug', default=False,
    #                     help='Verbose and debug output if set')

    args = parser.parse_args()
    return args


# taken from pysubs2:cli.py
def time(s):
    d = {}
    # all = re.findall(r"(\+?|-?|^$)(\d*\.?\d*)(ms|m|s|h)", s)
    sign = s[0] if s[0] == 'e' or s[0] == '+' else ''
    for v, k in re.findall(r"(\d*\.?\d*)(ms|m|s|h)", s):
        d[k] = float(v)
    return sign, make_time(**d)


def get_args_subs2cia():
    parser = argparse.ArgumentParser(description=f'subs2cia: Extract subtitled dialogue from audiovisual media for use '
                                                 f'in language acquisition')

    subparsers = parser.add_subparsers(title='subcommands', help="subcommand help", dest="command")
    parent_parser = argparse.ArgumentParser(add_help=False)

    parent_parser.add_argument('-v', '--verbose', action='store_true', dest='verbose', default=False, required=False,
                               help='Verbose output if set.')

    parent_parser.add_argument('-vv', '--debug', action='store_true', dest='debug', default=False, required=False,
                               help='Verbose and debug output if set')

    # todo: implement directory parsing
    # if directory is an input, batch mode must be specified
    parent_parser.add_argument('-i', '--inputs', metavar='<input files>', dest='infiles', default=None, required=False,
                               type=str, nargs='+',
                               help='Paths to input files or a single path to a directory of input files.')

    parent_parser.add_argument('-si', '--subtitle-index', metavar='<index>', dest='subtitle_stream_index', default=None,
                               type=int,
                               help='Force a certain subtitle stream to use. '
                                    'Takes precedence over --target-language option.'
                                    'If any input files are standalone subtitle files, they will be used first. '
                                    'Use --list-streams for a list of available streams and their indices.')

    parent_parser.add_argument('-ai', '--audio-index', metavar='<index>', dest='audio_stream_index', default=None,
                               type=int,
                               help='Force a certain subtitle audio to use. '
                                    'Takes precedence over --target-language option.'
                                    'If any input files are standalone audio files, they will be used first. '
                                    'Use --list-streams for a list of available streams and their indices.')

    parent_parser.add_argument('-b', '--batch', action='store_true', dest='batch', default=False,
                               help='If set, attempts to split input files into groups, one set of outputs per group. '
                                    'Groups are determined by file names. If two files share the same root name, such as '
                                    '"video0.mkv" and "video0.srt", then they are part of the same group. '
                                    'If file names contain a language code as a suffix, '
                                    'then the suffix will also be ignored (e.g. "video1.eng.flac" and "video1.ja.srt" '
                                    'will be grouped together under "video1")')

    parent_parser.add_argument('-u', '--dry-run', action='store_true', dest='dry_run', default=False,
                               help="If set, will analyze input files but won't demux or generate any output files")

    parent_parser.add_argument('-o', '--output-name', metavar='<name>', dest='outstem', default=None,
                               type=str,
                               help='Output file name to save to, without the extension '
                                    '(specify extension using -ae or -ve). '
                                    'By default, uses the file name of the first '
                                    'input file with the input extension removed '
                                    'and "condensed.{output extension} added. \n'
                                    'Ignored if batch mode is enabled.')

    parent_parser.add_argument('-d', '--output-dir', metavar='/path/to/directory', dest='outdir', default=None,
                               type=str,
                               help='Output directory to save to. Default is the directory the input files reside in.')

    parent_parser.add_argument('-ae', '--audio-extension', metavar='<audio extension>', dest='out_audioext',
                               default='mp3',
                               type=str,
                               help='Output audio extension to save as (without the dot). '
                                    'Default is mp3.')

    parent_parser.add_argument('-q', '--bitrate', metavar='<bitrate in kbps>', dest='bitrate', default=320,
                               type=int,
                               help='Output audio bitrate in kbps, lower bitrates result in smaller files and lower '
                                    'fidelity. Ignored if the output type is not mp3. Default is 320 kbps. '
                                    'Bitrates below 64 kbps are not recommended.')

    parent_parser.add_argument('-M', '--mono', dest='mono_channel', default=False,
                               action="store_true",
                               help='If set, mixes audio channels to a single channel, primarily to save space.')
    # todo: dot stripper, output naming
    parent_parser.add_argument('-m', '--gen-video', action='store_true', dest='condensed_video', default=False,
                               help='If set, generates condensed video along with condensed audio and subtitles. '
                                    'Subtitles are muxed in to video file. '
                                    'WARNING: VERY CPU INTENSIVE).')

    parent_parser.add_argument('--overwrite-on-demux', action='store_true', dest='demux_overwrite_existing',
                               default=False,
                               help='If set, will overwrite existing files when demuxing temporary files.')

    parent_parser.add_argument('--keep-temporaries', action='store_true', dest='keep_temporaries', default=False,
                               help='If set, will not delete any demuxed temporary files.')

    parent_parser.add_argument('--no-overwrite-on-generation', action='store_false',
                               dest='overwrite_existing_generated',
                               default=True,
                               help='If set, will not overwrite existing files when generating output media.')

    parent_parser.add_argument('-ni', '--ignore-none', action='store_true', dest='use_all_subs',
                               default=False,
                               help='If set, will not use internal heuristics to remove non-dialogue '
                                    'lines from the subtitle. Ignored if -R is set.')

    parent_parser.add_argument('-R', '--sub-regex-filter', metavar='<regular expression>', dest='subtitle_regex_filter',
                               default=None,
                               type=str,
                               help='For filtering non-dialogue subtitles. Lines that match given regex are IGNORED '
                                    'during subtitle processing and will not influence condensed audio or be included in output cards. '
                                    'Ignored lines may still be included in condensed subtitles if they overlap with non-ignored subtitles. '
                                    'This option will override the internal subs2cia non-dialogue filter.')

    # parent_parser.add_argument('-RR', '--sub-regex-substrfilter', metavar='<regular expression>',
    #                            dest='subtitle_regex_substrfilter', default=None,
    #                            type=str,
    #                            help='Searches subtitle lines and removes all substrings that match this regular '
    #                                 'expression. If the resulting subtitle line becomes empty, contains only spaces, '
    #                                 'or contains only punctuation as a result, the entire subtitle line is '
    #                                 'removed and will not be present in the output unless -RRnk is set. Internal '
    #                                 'dialogue detection method is still active unless -ni is set.')
    #
    # parent_parser.add_argument('-RRnk', '--sub-regex-substrfilter-nokeepchanges', action='store_true', dest='subtitle_regex_substrfilter_nokeep',
    #                            default=False,
    #                            help='If set, the modified subtitle text created by -RR will not be passed to the '
    #                                 'output subtitles/cards. Any empty filtered lines will be marked as non-dialogue '
    #                                 'instead of being removed entirely from the output. '
    #                                 'Useful for implementing custom dialogue-ignoring ')

    parent_parser.add_argument('-I', '--ignore-range', metavar="[prefix]timestamp", dest="ignore_range", default=None,
                               type=time, nargs=2, action="append",
                               help="Time range to ignore when condensing, specified using two timestamps. "
                                    "Useful for removing openings and endings of shows. \n"
                                    "Time formatting example: '2h30m2s100ms', '10m20s', etc. \n"
                                    "Subtitles that fall into an ignored range before padding are trimmed so that they "
                                    "do not overlap with the ignore range. "
                                    "Timestamps can measured from the start of the audio (no prefix), end of the audio (using the 'e' prefix), or relative to "
                                    "another timestamp (using the '+' prefix). "
                               # "Examples: \n"
                               #     "\t-I 1m 2m30s  # ignore subtitles that exist between 01:00 and 02:30\n"
                               #     "\t-I 1m12s +1m30s  # ignore subtitiles in the 1m30s after 01:12\n"
                               #     "\t-I -3m -1ms  # ignore subtitles that exist between 3 minutes from end of audio and 1 minute from end of audio"
                               #     "\t-I -2m +1m30s  # ignore subtitles that exist in the first 1m30s of the last 2 minutes of audio"
                                    "If batch mode is enabled, the same ranges are applied to ALL outputs."
                                    "Multiple ranges can be specified like so: -I 2m 3m30s -I 20m 21m. ")

    parent_parser.add_argument('-Ic', '--ignore-chapter', metavar="<chapter name>", dest="ignore_chapters",
                               default=None,
                               type=str, action="append",
                               help="Chapter titles to ignore, case sensitive. Can use -ls to determine chapter titles. "
                                    "Can be used in addition to --ignore-range to ignore sections of the stream.")

    parent_parser.add_argument('-p', '--padding', metavar='msecs', dest='padding', default=0,
                               type=int,
                               help='Adds this many milliseconds of audio before and after every subtitle. '
                                    'Overlaps with adjacent subtitles are merged automatically.')

    parent_parser.add_argument('-tl', '--target-language', metavar='ISO_code', dest='target_lang', default=None,
                               type=str,
                               help='If set, attempts to use audio and subtitle files that are in this language first. '
                                    'Input should be an ISO 639-3 language code.')

    parent_parser.add_argument('-ls', '--list-streams', dest='list_streams', action='store_true', default=False,
                               help='Lists all audio, subtitle, and video streams as well as chapters '
                                    'found in given input files and exits.')

    parent_parser.add_argument('--preset', metavar='preset#', dest='preset', type=int, default=None,
                               help='If set, uses a given preset. User arguments will override presets.')
    # todo: consider listing presets somewhere else
    parent_parser.add_argument('-lp', '--list-presets', dest='list_presets', action='store_true', default=False,
                               help='Lists all available built-in presets and exits.')

    # todo: consider depreciating this option, it's not really _that_ useful aside from user debugging
    parent_parser.add_argument('-a', '--absolute-paths', action='store_true', dest='absolute_paths', default=False,
                               help='Prints absolute paths from the root directory instead of given paths.')

    parent_parser.add_argument('-ma', '--interactive', action='store_true', dest='interactive', default=False,
                               help='If set, will enable interactive stream picking. Overrides -ai, -si, -tl. '
                                    'Also overrides -c in ')

    cia_parser = subparsers.add_parser('condense', parents=[parent_parser],
                                       help="Condense input audio into a single audio file")
    srs_parser = subparsers.add_parser('srs', parents=[parent_parser],
                                       help="Snip input media for import into spaced-repetition software")

    srs_parser.add_argument('-N', '--normalize', action='store_true', dest='normalize_audio', default=False,
                            help="If set, normalizes volume of audio clips to the same loudness. YMMV.")

    cia_parser.add_argument('-t', '--threshold', metavar='msecs', dest='threshold', default=0,
                            type=int,
                            help="If two subtitles start and end within (threshold + 2*padding) "
                                 "milliseconds of each other, they will be merged. "
                                 "Useful for preserving silences between "
                                 "subtitle lines.")

    cia_parser.add_argument('-r', '--partition', metavar='secs', dest='partition', default=0,
                            type=int,
                            help="If set, attempts to partition the input audio into "
                                 "seperate blocks of this size seconds BEFORE condensing. Partitions and splits respect "
                                 "subtitle boundaries and will not split a single subtitle across two output files."
                                 " 0 partition length is ignored. For example, if the partition size is 60 seconds and the "
                                 "input media is 180 seconds long, then there will be three output files. The first output "
                                 "file will contain condensed media from the first 60 seconds of the source material, "
                                 "the second output file will contain the next 60 seconds of input media, and so on.")

    cia_parser.add_argument('-s', '--split', metavar='secs', dest='split', default=0,
                            type=int,
                            help="If set, attempts to split the condensed audio into "
                                 "seperate blocks of this size AFTER condensing. 0 "
                                 "split length is ignored. "
                                 "Partitions and splits respect subtitle boundaries and will not split a single subtitle "
                                 "across two output files. "
                                 "Done within a partition. For example, say the split length is "
                                 "60 seconds and the condensed audio length of a input partition is 150 seconds. "
                                 "The output file will be split into three files, the first two ~60 seconds long and the "
                                 "last ~30 seconds long.")

    cia_parser.add_argument('-c', '--minimum-compression-ratio', metavar='<ratio>', dest='minimum_compression_ratio',
                            default=0.2, type=float,
                            help="Will only generate from subtitle files that are this fraction long of the selected audio "
                                 "file. Default is 0.2, meaning the output condensed file must be at least 20 percent as long as "
                                 "the chosen audio stream. If the output doesn't reach this minimum, then a different "
                                 "subtitle file will be chosen, if available. Used for ignoring subtitles that contain only "
                                 "signs and songs.")

    cia_parser.add_argument('--no-gen-subtitle', action='store_true', dest='no_condensed_subtitles', default=False,
                            help="If set, won't output a condensed subtitle file. Useful for reducing file clutter.")

    args = parser.parse_args()

    if len(sys.argv) == 1:
        parser.print_help()
        sys.exit(0)

    return args



from subs2cia.sources import Stream
from subs2cia.ffmpeg_tools import ffmpeg_demux, ffmpeg_trim_audio_clip_atrim_encode, ffmpeg_get_frame_fast

import logging
import pysubs2 as ps2  # for reading in subtitles
from pathlib import Path
from datetime import timedelta
import ffmpeg
import re
import copy
import sys
import unicodedata

codepoints = range(sys.maxunicode + 1)
punct_tbl = dict.fromkeys(i for i in range(sys.maxunicode)
                      if unicodedata.category(chr(i)).startswith('P') or unicodedata.category(chr(i)).startswith('S'))

from typing import List, Union


def overlap_range(range1: List[int], range2: List[int]):
    assert len(range1) == len(range2) == 2
    if range1[0] < range2[0] < range1[1] or range1[0] < range2[1] < range1[1] or range2[0] < range1[0] < range2[1] or \
        range2[0] < range1[1] < range2[1]:
        return True
    return False


def ssaevent_trim(event: ps2.SSAEvent, ir: List[int]):
    r"""
    Given an event that overlaps with ir
    if subtitle is only on one side of IR
        trim subtitle to not overlap IR
    if subtitle is on both sides of IR
        split subtitle into two and trim each
    if subtitle is entirely inside IR
        drop it
    :param event: SSAEvent to trim/split+trim/drop
    :param ir: Ignore range, two integers representing milliseconds
    :return: List of trimmed events. There may be zero, one, or two events returned
    """

    assert len(ir) == 2
    trimmed = []
    if ir[0] <= event.start < ir[1] and ir[0] < event.end <= ir[1]:
        # event falls completely inside an ignore range
        return trimmed
    if ir[0] <= event.start < ir[1] and ir[1] < event.end:
        # event starts inside ignore range, ends outside of it
        event.start = ir[1]
        trimmed.append(event)
        return trimmed
    if event.start < ir[0] and ir[0] < event.end <= ir[1]:
        # event starts outside of ignore range, ends inside of it
        event.end = ir[0]
        trimmed.append(event)
        return trimmed
    if event.start < ir[0] and ir[1] < event.end:
        # split into two pieces
        event2 = copy.deepcopy(event)
        event.end = ir[0]
        trimmed.append(event)
        event2.start = ir[1]
        trimmed.append(event2)
        return trimmed
    # should never get here
    assert False


def ignore_nibble(ignore_ranges: List[List[int]], e: ps2.SSAEvent):
    r"""
    Given a set of ignore_ranges and a SSAEvent, determine if the event falls in an IR and if it does, trim it
    :param ignore_ranges: List of IRs
    :param e: SSAEvent of interest
    :return: List of SSAEvents
    """
    for ir in ignore_ranges:
        assert len(ir) == 2
        if overlap_range(ir, [e.start, e.end]):
            trimmed = ssaevent_trim(e, ir)
            # assert(trimmed is not None)
            return trimmed
    return [e]


class SubGroup:
    def __init__(self, events: [ps2.SSAEvent], ephemeral: bool, threshold: int, padding: int):
        self.contains_only_ephemeral = ephemeral  # if true, won't affect merging/splitting
        self.events = events
        self.ephemeral_events = []  # not empty only when mixing ephemeral and dialogue events

        self.threshold = threshold
        self.padding = padding

    @property
    def events_start(self):
        lowest = float('inf')
        for e in self.events:  # there may be 0 events
            if e.start < lowest:
                lowest = e.start
        return lowest

    @property
    def events_end(self):
        highest = 0
        for e in self.events:  # there may be 0 events
            if e.end > highest:
                highest = e.end
        return highest

    @property
    def group_range(self):
        r"""
        Subtitle group start, end with padding
        :return: [range_start, range_end]
        """
        if self.contains_only_ephemeral:
            return [self.events_start, self.events_end]
        grange = self.padding
        return [self.events_start - grange if self.events_start - grange > 0 else 0, self.events_end + grange]

    @property
    def group_limits(self):
        r"""
        Subtitle group start/end extended with padding and threshold
        :return: [limit_start, limit_end]
        """
        if self.contains_only_ephemeral:
            return [self.events_start, self.events_end]
        limit = self.threshold/2 + self.padding  # divide by two: threshold is distance to next group
        return [self.events_start - limit if self.events_start - limit > 0 else 0, self.events_end + limit]

    def __repr__(self):
        s = f"<SubGroup |{self.group_limits[0]} {self.group_range[0]} {(self.events_start, self.events_end)} {self.group_range[1]} {self.group_limits[1]}|>"
        return s


class SubtitleManipulator:
    def __init__(self, subpath: Path, threshold: int, padding: int, ignore_range: Union[List[List[int]], None], audio_length: int):
        r"""
        Class for subtitle manipulation
        :param subpath: Path to pysubs2-compatible subtitle file
        :param threshold: in milliseconds
        :param padding: in milliseconds
        :param ignore_range: list of ranges. each range is a list of two tuples. each tuple contains two values, a "sign" and a "duration".
        :param audio_length: How long the audio is in milliseconds.
        """
        self.subpath = subpath
        self.ssadata = None
        self.condensed_ssadata = None
        self.ssa_events = None
        self.groups = None
        self.ephemeral = None

        self.threshold = threshold
        self.padding = padding

        self.audio_length = audio_length

        self.ignore_range = None
        if ignore_range is not None:
            self.ignore_range = []
            for irange in ignore_range:
                assert(len(irange) == 2)
                to_append = []
                for idx, (sign, duration) in enumerate(irange):
                    if sign == "+":
                        if idx == 0:
                            raise AssertionError("Can't have duration '+' as first range value, "
                                                 "must be second range value "
                                                 "(e.g. -I 1m +1m30s)")
                        to_append.append(to_append[0] + duration)
                    elif sign == "e":
                        to_append.append(int(self.audio_length - duration))
                    elif sign == "":
                        to_append.append(duration)
                    else:
                        # shouldn't get here, re.findall should strip anything unexpected out
                        raise AssertionError(f"SubtitleManipulator received unexpected sign ({sign})")
                if not to_append[0] < to_append[1]:
                    raise AssertionError(f"Ignore range '{irange[0][0]}{irange[0][1]}ms {irange[1][0]}{irange[1][1]}ms'"
                                         f" is invalid: end of range "
                                         f"({to_append[1]}ms) is before start of range ({to_append[0]}ms)")
                self.ignore_range.append(to_append)

    def load(self, include_all: bool, regex: str, substrreplace_regex: str, substrreplace_nokeepchanges: bool):
        if not self.subpath.exists():
            logging.warning(f"Subtitle file {self.subpath} does not exist")
            return

        logging.debug(f"Loading subtitles at {self.subpath}")
        try:
            self.ssadata = ps2.load(str(self.subpath))
        except (ps2.FormatAutodetectionError) as e:
            # retry by forcing format
            logger = logging.getLogger(__name__)
            logger.exception(e)
            logging.warning(f"Subtitle file format not recognized by pysubs2, will try forcing extension {self.subpath.suffix[1:]} as format ({self.subpath})")
            try:
                self.ssadata = ps2.load(str(self.subpath), format_=self.subpath.suffix[1:])
            except Exception as e:
                logger = logging.getLogger(__name__)
                logger.exception(e)
                logging.warning(f"pysubs2 enountered error loading subtitle file {self.subpath}")
                self.ssadata = None
                return
        except (ValueError, AttributeError) as e:
            logger = logging.getLogger(__name__)
            logger.exception(e)
            logging.warning(f"pysubs2 enountered error loading subtitle file {self.subpath}")
            self.ssadata = None
            return
        logging.debug(f"Loaded {self.ssadata}")

        self.ssa_events = self.ssadata.events
        self.ssa_events.sort(key=lambda x: x.start)

        pool = self.ssa_events
        self.groups = []
        while len(pool) > 0:
            e = pool.pop(0)
            ignored = False
            if substrreplace_regex:
                new_subtitle_line = re.sub(substrreplace_regex, '', e.plaintext)
                if new_subtitle_line == e.plaintext:
                    pass
                else:
                    logging.debug(f"substrfilter: modified line:\nold line: {e.plaintext}\nnew line: {new_subtitle_line}")
                    # is the line now empty or contains only spaces or punctuation/symbols
                    if len(new_subtitle_line) == 0 or \
                       new_subtitle_line.isspace() or \
                       len(new_subtitle_line.translate(punct_tbl)) == 0:
                        if substrreplace_nokeepchanges:
                            # mark as non-dialogue
                            ignored = True
                        else:
                            # remove entirely
                            continue
                    # do we want to use the original e.plaintext or the stripped and non-empty new_subtitle_line?
                    if not substrreplace_nokeepchanges:
                        e.plaintext = new_subtitle_line
            if self.ignore_range is not None:
                if any([overlap_range(ir, [e.start, e.end]) for ir in self.ignore_range]):
                    trimmed = ignore_nibble(self.ignore_range, e)
                    pool = trimmed + pool
                    continue  # trimmed events may still overlap other ranges, will need to retest
            self.groups.append(SubGroup([e], ephemeral=ignored or not is_dialogue(e, include_all, regex),
                                        threshold=self.threshold,
                                        padding=self.padding))

    def merge_groups(self):
        merged = []
        self.ephemeral = []
        for group in self.groups:
            # ephermal groups are not merged with any other groups so they can be dealt with seperately
            if group.contains_only_ephemeral:
                self.ephemeral.append(group)
                continue
            if len(merged) == 0:  # first group
                merged.append(group)
                continue
            if merged[-1].group_limits[1] > group.group_limits[0]:
                merged[-1].events += group.events
            else:
                merged.append(group)
        self.groups = merged
        logging.debug(f"Merged groups {merged}")
        # add ephemeral groups back into rest of groups
        for egroup in self.ephemeral:  # assuming each egroup contains one ssaevent
            for group in self.groups:
                # if egroup is valid inside group range
                    # create a new ssaevent from egroup that fits inside group
                if group.group_range[0] < egroup.events_start < group.group_range[1] or \
                group.group_range[0] < egroup.events_end < group.group_range[1]:
                    new_event = egroup.events[0].copy()
                    new_event.start = new_event.start if new_event.start > group.group_range[0] else group.group_range[0]
                    new_event.end = new_event.end if new_event.end < group.group_range[1] else group.group_range[1]
                    group.ephemeral_events.append(new_event)
        logging.debug("Inserted ephemeral events")

    def get_times(self):
        # returns list of tuples of group ranges
        times = []
        for g in self.groups:
            times.append(g.group_range)
        return times

    def condense(self):
        r"""
        laststart = 0
        for each group g
            shift all event times in g back by g.group_range[0] - laststart milliseconds
        :return:
        """
        laststart = 0
        groups = copy.deepcopy(self.groups)
        for g in groups:
            shift = g.group_range[0] - laststart
            for e in g.events:
                e.start -= shift
                e.end -= shift
            for e in g.ephemeral_events:
                e.start -= shift
                e.end -= shift
            laststart = g.group_range[1]
        logging.debug("Shifted subtitle groups")
        # extract shifted SSAevents
        condensed_events = []
        for g in groups:
            for e in g.events:
                condensed_events.append(e)
            for e in g.ephemeral_events:
                condensed_events.append(e)
        self.condensed_ssadata = copy.copy(self.ssadata)
        self.condensed_ssadata.events = condensed_events


def decide_partitions(sub_times, partition=0):
    if partition == 0:
        return [(0, len(sub_times)), ]

    partitions = list()
    current_partition = 0
    start = 0
    idx = 0
    for idx, t in enumerate(sub_times):
        current_partition_boundary = (current_partition + 1) * partition
        if t[1] > current_partition_boundary:
            if abs(t[1] - current_partition_boundary) < abs(sub_times[idx - 1][1] - current_partition_boundary):
                end = idx + 1
            else:
                end = idx
            partitions.append((start, end))
            current_partition += 1
            start = end
    end = idx + 1
    partitions.append((start, end))
    return partitions


def split_times(sub_times, partition_indicies, splitsize=0):
    partition_times = sub_times[partition_indicies[0]:partition_indicies[1]]
    if splitsize == 0:
        return [partition_times, ]

    prev = 0
    for idx, time in enumerate(sub_times):
        time.append(time[1] - time[0] + prev)
        prev = time[2]

    start = 0
    idx = 0
    splits = list()
    current_split = 0
    for idx, t in enumerate(partition_times):
        current_partition_boundary = (current_split + 1) * splitsize
        if t[2] > current_partition_boundary:
            if abs(t[2] - current_partition_boundary) < abs(partition_times[idx - 1][2] - current_partition_boundary):
                end = idx + 1
            else:
                end = idx
            splits.append((start, end))
            current_split += 1
            start = end
    end = idx + 1
    splits.append((start, end))
    splittimes = list()
    for split in splits:
        tmp = partition_times[split[0]:split[1]]
        tmp = [[x[0], x[1]] for x in tmp]
        splittimes.append(tmp)
    return splittimes


def partition_and_split(sub_times, partition_size=0, split_size=0):
    # returns a list tuple index pairs in [start, end) format for sub_times
    partitions = decide_partitions(sub_times,
                                   partition=partition_size)
    """
    partition1
        split1
            time1
                start
                end
            time2
                ...
        split2
            time1
        split3
        split4...
    partition2
        split1
            time1
            time2...
    """
    divided_times = list()
    for p, partition_size in enumerate(partitions):
        # returns list of list of times in milliseconds, can feed into condense_audio
        times = split_times(sub_times, partition_size,
                            splitsize=split_size)
        divided_times.append(times)

    return divided_times


# given raw subtitle timing data, merge overlaps and perform padding and merging
# DEPRECIATED
def merge_times(times: [], threshold=0, padding=0):
    # padding: time to add to the beginning and end of each subtitle. Resulting overlaps are removed.
    # threshold: if two subtitles are within this distance apart, they will be merged in the audio track
    times.sort(key=lambda x: x[0])  # sort by first value in tuple

    initial = len(times)
    if padding > 0:
        for time in times:
            time[0] -= padding
            time[1] += padding

    # even if threshold is 0 any added padding may cause overlaps, which the code below will also merge
    idx = 0
    while idx < len(times) - 1:
        if times[idx + 1][0] < times[idx][1] + threshold:
            # if the next subtitle starts within 3 seconds of the current subtitle finishing, merge them
            times[idx][1] = times[idx + 1][1]
            del times[idx + 1]
            if idx < len(times) - 1:
                continue
        idx += 1

    logging.info(f"Merged {initial} subtitles into {len(times)} audio snippets\n")
    logging.debug(f"Merged times: {times}")
    return times


# examine a subtitle's text and determine if it contains text or just signs/song styling
# could be much more robust, by including regexes
def is_dialogue(line, include_all=False, regex=None):
    if regex is not None:
        p = re.compile(regex)
        logging.debug(f"using regex {regex} for subtitle search")
        m = p.findall(line.text)
        if len(m) >= 1:
            logging.debug(f'Regex {regex} found match(s) in "{line.text}": {m}')
            return False
        return True
    if include_all:  # ignore filtering
        return True
    if line.type != "Dialogue":
        return False
    if len(line.text) == 0:
        return False
    if '{' == line.text[0]:
        return False
    if '（' == line.text[0] and '）' == line.text[-1]:
        return False

    # list of characters that, if present anywhere in the subtitle text, means that the subtitle is not dialogue
    globally_invalid = ["♪", "♪"]
    if any(invalid in line.text for invalid in globally_invalid):
        return False
    return True


# given a path to a subtitle file, load it in and strip it of non-dialogue text
# keyword arguments are filter options that are passed to is_spoken_dialogue for filtering configuration
# DEPRECIATED
def load_subtitle_times(subfile: Path, include_all_lines=False):
    # if dry_run:
    #     print("dry_run: will load", subfile)
    #     if os.path.isfile(subfile) is False:
    #         print(f"note: {subfile} doesn't exist yet, skipping subtitle loading")
    #         return [[0, 0], ]  # might just be a demuxed file that hasn't been demuxed yet
    logging.info(f"loading {subfile}")
    try:
        subs = ps2.load(str(subfile))
    except (ValueError, AttributeError) as e:
        logger = logging.getLogger(__name__)
        logger.exception(e)
        logging.warning(f"pysub2 enountered error loading subtitle file {subfile}, trying a different subtitle file...")
        return None
    logging.info(subs)

    times = list()
    count = 0
    for idx, line in enumerate(subs):
        # print(line)
        if is_dialogue(line, include_all=include_all_lines):
            # print(line)
            times.append([line.start, line.end])
        else:
            logging.debug(f"ignoring {line.text}, probably not spoken dialogue")
            count += 1
            pass
    logging.info(f"ignored {count} lines")  # number of subtitles that looked like they didn't contain dialogue

    # print(times)
    if len(times) == 0:
        logging.warning(f"warning: got 0 dialogue lines from subtitle file {subfile}")
        return None
    return times


def get_partitioned_and_split_times_duration(sub_times):
    r"""
    Calculates the total amount of condensed audio to be generated from the subtitle times, in milliseconds
    :param sub_times:
    :return: time in milliseconds
    """
    total = list()
    for split in sub_times:
        for partition in split:
            for x1, x2 in partition:
                total.append(x2 - x1)
    subs_total = sum(total)
    return subs_total


# don't use: not reliable
# def get_audiostream_duration(stream_info):
#     sps = int(
#         stream_info['time_base'].split('/')[1])  # audio samples per second, inverse of sampling frequency
#     samples = stream_info['duration_ts']  # total samples in audio track
#     audio_total = samples / sps * 1000
#     return audio_total


def get_audiofile_duration(audiofile: Path):
    stream_idx = 0
    audio_info = ffmpeg.probe(str(audiofile), cmd='ffprobe')
    sps = int(
        audio_info['streams'][stream_idx]['time_base'].split('/')[1])  # audio samples per second, inverse of sampling frequency
    samples = audio_info['streams'][stream_idx]['duration_ts']  # total samples in audio track
    audio_total = samples / sps * 1000
    return audio_total


def get_compression_ratio(sub_times, audiofile: Path, verbose=True):
    # audio_info = ffmpeg.probe(str(audiofile), cmd='ffprobe')
    # sps = int(
    #     audio_info['streams'][0]['time_base'].split('/')[1])  # audio samples per second, inverse of sampling freq
    # samples = audio_info['streams'][0]['duration_ts']  # total samples in audio track
    # audio_total = samples / sps * 1000
    audio_total = get_audiofile_duration(audiofile)
    # total = list()
    # for split in sub_times:
    #     for partition in split:
    #         for x1, x2 in partition:
    #             total.append(x2-x1)
    # subs_total = sum(total)
    subs_total = get_partitioned_and_split_times_duration(sub_times)
    if verbose:
        logging.info(f"will condense {str(timedelta(milliseconds=audio_total)).split('.')[0]} of source audio to "
                     f"{str(timedelta(milliseconds=subs_total)).split('.')[0]} "
                     f"({round(subs_total / audio_total * 100, 1)}% compression ratio) of condensed audio")
    return subs_total / audio_total



from subs2cia.sources import AVSFile
from subs2cia.pickers import picker
from subs2cia.sources import Stream, get_and_partition_streams
import subs2cia.subtools as subtools
from subs2cia.ffmpeg_tools import export_condensed_audio, export_condensed_video

from typing import List, Union, Dict
from collections import defaultdict
from pathlib import Path
from pprint import pprint
import pysubs2 as ps2

import logging


def picked_sources_are_insufficient(d: dict):
    for k in d:
        if d[k] == 'retry':
            return True
    if d['subtitle'] is None:
        return True
    if d['audio'] is None:
        return True
    return False


def insufficient_source_streams(d: dict):
    if len(d['subtitle']) == 0:
        return True
    if len(d['audio']) == 0:
        return True
    return False


def chapter_timestamps(sourcefile: AVSFile, ignore_chapters: List[str]):
    if len(ignore_chapters) == 0:
        return []

    chapters = sourcefile.info['chapters'] or []

    if len(chapters) == 0:
        return []

    chapters_by_title = {c['tags']['title']: c for c in chapters}
    timestamps = []

    for title in ignore_chapters:
        if title in chapters_by_title:
            chapter = chapters_by_title[title]
            timestamps.append([('', 1000 * int(float((chapter['start_time'])))), ('', 1000 * int(float(chapter['end_time'])))])
        else:
            logging.warning(f"Chapter '{title}' was specified to be ignored, but it was not found")

    return timestamps


def interactive_picker(sources: List[AVSFile], partitioned_streams: Dict[str, Stream], media_type: str):
    print("Input files:")
    for avsf in sources:
        print(f"\t{avsf}")
    print(f"Found the following {media_type} streams:")
    for idx, stream in enumerate(partitioned_streams[media_type]):
        desc_str = ''
        if "codec_name" in stream.stream_info:
            desc_str = desc_str + "codec: " + stream.stream_info['codec_name'] + ", "
        if media_type == 'video':
            if 'width' in stream.stream_info and 'height' in stream.stream_info:
                desc_str = desc_str + f"{stream.stream_info['width']}x{stream.stream_info['height']}, "
        if media_type == 'audio':
            if "tags" in stream.stream_info:
                tags = stream.stream_info['tags']
                if "language" in tags:
                    desc_str = desc_str + "lang_code: " + tags['language'] + ", "
                if "title" in tags:
                    desc_str = desc_str + "title: " + tags['title'] + ", "
        if desc_str == '':
            desc_str = f"\tStream {idx: 3}: no information found, "
        else:
            desc_str = f"\tStream {idx: 3}: {desc_str}"
        desc_str += f"file: {str(stream.file)}"
        print(desc_str)
    print("")
    idx = int(input("Which stream to use?"))
    return partitioned_streams[media_type][idx]


class Common:
    def __init__(self, sources: List[AVSFile], outdir: Union[Path, None], outstem: Union[str, None],
                 condensed_video: bool, padding: int,
                 demux_overwrite_existing: bool, overwrite_existing_generated: bool,
                 keep_temporaries: bool, target_lang: str, out_audioext: str,
                 use_all_subs: bool, subtitle_regex_filter: str, subtitle_regex_substrfilter: str,
                 subtitle_regex_substrfilter_nokeep: bool, audio_stream_index: int, subtitle_stream_index: int,
                 ignore_range: Union[List[List[int]], None], ignore_chapters: Union[List[str], None],
                 bitrate: Union[int, None], mono_channel: bool, interactive: bool):
        if outdir is None:
            self.outdir = sources[0].filepath.parent
        else:
            self.outdir = outdir

        self.sources = sources
        if outstem is not None:
            self.outstem = outstem
        else:
            self.outstem = sources[0].filepath.stem

        self.partitioned_streams = defaultdict(list)

        self.picked_streams = {
            'audio': None,
            'subtitle': None,
            'video': None
        }

        self.pickers = {
            'audio': None,
            'subtitle': None,
            'video': None
        }

        self.quality = bitrate
        self.to_mono = mono_channel

        self.condensed_video = condensed_video
        self.padding = padding
        self.demux_overwrite_existing = demux_overwrite_existing
        self.overwrite_existing_generated = overwrite_existing_generated
        self.keep_temporaries = keep_temporaries
        self.target_lang = target_lang

        self.out_audioext = out_audioext
        self.out_videoext = '.mp4'


        self.use_all_subs = use_all_subs
        self.subtitle_regex_filter = subtitle_regex_filter
        self.subtitle_regex_substrfilter = subtitle_regex_substrfilter
        self.subtitle_regex_substrfilter_nokeep = subtitle_regex_substrfilter_nokeep
        self.audio_stream_index = audio_stream_index
        self.subtitle_stream_index = subtitle_stream_index
        # todo: verify ignore_range, make sure start is after end
        self.ignore_range = ignore_range
        self.ignore_chapters = ignore_chapters

        self.insufficient = False

        self.interactive = interactive

    # go through source files and count how many subtitle and audio streams we have
    def get_and_partition_streams(self):
        self.partitioned_streams = get_and_partition_streams(self.sources)
        # for sourcefile in self.sources:
        #     if sourcefile.type == 'video':
        #         # dig into streams
        #         for idx, stream_info in enumerate(sourcefile.info['streams']):
        #             stype = stream_info['codec_type']
        #             self.partitioned_streams[stype].append(Stream(file=sourcefile, type=stype,
        #                                                           index=idx, stream_info=stream_info))
        #         continue
        #     self.partitioned_streams[sourcefile.type].append(Stream(file=sourcefile, type=sourcefile.type,
        #                                                             stream_info=sourcefile.info,
        #                                                             index=None))
        #     # for stream in sourcefile
        # for k in self.partitioned_streams:
        #     logging.info(f"Found {len(self.partitioned_streams[k])} {k} input streams")
        #     # logging.debug(f"Streams found: {self.partitioned_streams[k]}")

    def initialize_pickers(self):
        for k in self.pickers:
            idx = None
            if k == 'audio':
                idx = self.audio_stream_index
            if k == 'subtitle':
                idx = self.subtitle_stream_index
            self.pickers[k] = picker(self.partitioned_streams[k], target_lang=self.target_lang, forced_stream=idx)

    def list_streams(self):
        source_string = ', '.join([str(s) for s in self.sources])
        print(f"Listing streams found in {source_string}")
        for k in ['subtitle', 'audio', 'video']:
            print(f"Available {k} streams:")
            for idx, stream in enumerate(self.partitioned_streams[k]):
                desc_str = ''
                if "codec_name" in stream.stream_info:
                    desc_str = desc_str + "codec: " + stream.stream_info['codec_name'] + ", "
                if "tags" in stream.stream_info:
                    tags = stream.stream_info['tags']
                    if "language" in tags:
                        desc_str = desc_str + "lang_code: " + tags['language'] + ", "
                    if "title" in tags:
                        desc_str = desc_str + "title: " + tags['title'] + ", "
                desc_str = desc_str + f"[{stream.file.filepath}]"
                if desc_str == '':
                    desc_str = f"Stream {idx: 3}: no information found"
                else:
                    desc_str = f"Stream {idx: 3}: {desc_str}"
                print(desc_str)
            print("")

        print(f"Available chapters:")
        for source in self.sources:
            if source.type != 'video':
                continue
            if 'chapters' not in source.info or len(source.info['chapters']) == 0:
                continue
            chapters = source.info['chapters']
            chapters_by_title = {c['tags']['title']: c for c in chapters}
            # pprint(chapters_by_title)
            for k in chapters_by_title:
                start = ps2.time.ms_to_str(float(chapters_by_title[k]['start_time']) * 1000)
                end = ps2.time.ms_to_str(float(chapters_by_title[k]['end_time']) * 1000)
                print(f'{start} - {end} "{k}"')
        print("\n")

    def choose_audio(self, interactive: bool):
        r"""
        Picks an audio stream to use. If there are multiple audio streams, prioritizes the following:
            * CLI-specified stream index
            * target language
            * order of appearance
        :param interactive: if true, prompts user about which audio stream to use
        :return:
        """
        if len(self.partitioned_streams['audio']) == 0:
            logging.warning(f"Couldn't find audio streams in input files")
            return
        if interactive and len(self.partitioned_streams['audio']) > 1:
            self.picked_streams['audio'] = interactive_picker(self.sources, self.partitioned_streams, 'audio')
            return

        k = 'audio'
        while self.picked_streams[k] is None:
            try:
                self.picked_streams[k] = next(self.pickers[k])
            except StopIteration as s:
                logging.critical("Inputs don't contain usable audio")
                self.insufficient = True
                return
            afile = self.picked_streams[k].demux(overwrite_existing=self.demux_overwrite_existing)
            if afile is None:
                logging.warning(f"Error while demuxing {self.picked_streams[k]}")
                self.picked_streams[k] = None

    def choose_subtitle(self, interactive: bool):
        # pass
        raise NotImplementedError("Child classes must implement choose_subtitles")

    def choose_video(self, interactive: bool):
        if len(self.partitioned_streams['video']) == 0:
            logging.info(f"Couldn't find video streams in input files")
            return
        if interactive and len(self.partitioned_streams['audio']) > 1:
            self.picked_streams['audio'] = interactive_picker(self.sources, self.partitioned_streams, 'video')
            return
        k = 'video'
        while self.picked_streams[k] is None:
            try:
                self.picked_streams[k] = next(self.pickers[k])
            except StopIteration as s:
                logging.critical("Inputs don't contain usable audio")
                self.insufficient = True
                return
            # afile = self.picked_streams[k].demux(overwrite_existing=self.demux_overwrite_existing)
            # if afile is None:
            #     logging.warning(f"Error while demuxing {self.picked_streams[k]}")
            #     self.picked_streams[k] = None

    def choose_streams(self):
        if insufficient_source_streams(self.partitioned_streams):
            logging.error(f"Not enough input sources to generate condensed output for output stem {self.outstem} "
                          f"(missing audio and/or subtitles)")
            self.insufficient = True
            return
        self.choose_audio(interactive=self.interactive)
        self.choose_subtitle(interactive=self.interactive)
        self.choose_video(interactive=self.interactive)

        logging.info(f"Picked audio stream:    {self.picked_streams['audio']}")
        logging.info(f"Picked subtitle stream: {self.picked_streams['subtitle']}")
        logging.info(f"Picked video stream:    {self.picked_streams['video']}")

    def choose_streams_old(self):
        if insufficient_source_streams(self.partitioned_streams):
            logging.error(f"Not enough input sources to generate condensed output for output stem {self.outstem}")
            self.insufficient = True
            return
        while picked_sources_are_insufficient(self.picked_streams):
            for k in ['audio', 'video', 'subtitle']:
                if len(self.partitioned_streams[k]) == 0:
                    logging.debug(f"no input streams of type {k}")
                    continue
                if self.picked_streams[k] is None:
                    try:
                        self.picked_streams[k] = next(self.pickers[k])
                    except StopIteration as s:
                        logging.critical("Input streams for this input group are invalid for condensing "
                                         "(missing audio or subtitles)")
                        self.insufficient = True
                        return
            for k in ['audio', 'subtitle', 'video']:
                # validate picked streams


                if k == 'audio':
                    afile = self.picked_streams[k].demux(overwrite_existing=self.demux_overwrite_existing)
                    if afile is None:
                        self.picked_streams[k] = None

                # todo: maybe move this back into Condense, split/threshold is a cia thing, not a srs thing,
                #  or have seperate functions for video, audio, and subtitles
                if k == 'subtitle':
                    subfile = self.picked_streams[k].demux(
                        overwrite_existing=self.demux_overwrite_existing)  # type AVSFile
                    if subfile is None:
                        self.picked_streams[k] = None
                        continue
                    # times = subtools.load_subtitle_times(subfile.filepath, include_all_lines=self.use_all_subs)
                    audiolength = subtools.get_audiofile_duration(
                        self.picked_streams['audio'].demux_file.filepath)
                    subdata = subtools.SubtitleManipulator(subfile.filepath,
                                                           threshold=self.threshold, padding=self.padding,
                                                           ignore_range=self.ignore_range, audio_length=audiolength)
                    subdata.load(include_all=self.use_all_subs, regex=self.subtitle_regex_filter)
                    if subdata.ssadata is None:
                        self.picked_streams[k] = None
                        continue
                    if self.picked_streams['audio'] is None:
                        # can't verify subtitle validity until audio candidate is found
                        continue
                    # times = subtools.merge_times(times, threshold=self.threshold, padding=self.padding)
                    subdata.merge_groups()
                    times = subdata.get_times()
                    ps_times = subtools.partition_and_split(times, self.partition, self.split)

                    sublength = subtools.get_partitioned_and_split_times_duration(ps_times)

                    compression_ratio = sublength / audiolength
                    if compression_ratio < self.minimum_compression_ratio:
                        logging.info(f"got compression ratio of {compression_ratio}, which is smaller than the minimum"
                                     f" ratio of {self.minimum_compression_ratio}, retrying wth different subtitle "
                                     f"file. If too many subtitles are being ignored, try -ni or -R. If the minimum "
                                     f"ratio is too high, try -c.")
                        self.picked_streams[k] = None
                        continue
                    self.dialogue_times = subtools.partition_and_split(sub_times=times,
                                                                       partition_size=1000 * self.partition,
                                                                       split_size=1000 * self.split)
                if k == 'video':
                    pass
        logging.info(f"Picked {self.picked_streams['audio']} to use for condensing")
        logging.info(f"Picked {self.picked_streams['video']} to use for condensing")
        logging.info(f"Picked {self.picked_streams['subtitle']} to use for condensing")

    def cleanup(self):
        if self.keep_temporaries:
            return
        for k in ['audio', 'video', 'subtitle']:
            if len(self.partitioned_streams) == 0:
                continue
            for s in self.partitioned_streams[k]:
                s.cleanup_demux()
